/*******************************************************************************
 * Copyright (c) 2001, 2007 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.soda.sat.core.internal.framework.bundle;

import org.eclipse.soda.sat.core.internal.nls.MessageFormatter;
import org.eclipse.soda.sat.core.internal.nls.Messages;
import org.eclipse.soda.sat.core.internal.util.Assert;
import org.eclipse.soda.sat.core.util.LogUtility;
import org.osgi.framework.BundleContext;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.FrameworkListener;

/**
 * FrameworkManager.java
 */
class FrameworkManager extends Object {
	//
	// Static Fields
	//

	// Externalized String Keys
	private static final String FRAMEWORK_ERROR_HAS_OCCURRED_KEY = "FrameworkManager.FrameworkErrorOccurred";  //$NON-NLS-1$
	private static final String FREE_MEMORY_AFTER_GC_KEY = "FrameworkManager.FreeMemoryAfterGc";  //$NON-NLS-1$
	private static final String FREE_MEMORY_BEFORE_GC_KEY = "FrameworkManager.FreeMemoryBeforeGc";  //$NON-NLS-1$
	private static final String FREE_MEMORY_DELTA_KEY = "FrameworkManager.FreeMemoryDelta";  //$NON-NLS-1$
	private static final String TOTAL_MEMORY_KEY = "FrameworkManager.TotalMemory";  //$NON-NLS-1$
	private static final String UNKNOWN_BUNDLE_CONTEXT_KEY = "Common.UnknownBundleContext";  //$NON-NLS-1$
	private static final String UNKNOWN_FRAMEWORK_EVENT_KEY = "FrameworkManager.UnknownFrameworkEvent";  //$NON-NLS-1$

	// Singleton
	private static final FrameworkManager INSTANCE = new FrameworkManager();

	//
	// Static Methods
	//

	/**
	 * Answers the <code>FrameworkManager</code> singleton instance.
	 *
	 * @return FrameworkManager
	 */
	static FrameworkManager getInstance() {
		return FrameworkManager.INSTANCE;
	}

	//
	// Instance Fields
	//

	private BundleContext bundleContext;
	private FrameworkListener frameworkListener;

	//
	// Constructors
	//

	/**
	 * Private constructor.
	 */
	private FrameworkManager() {
		super();
		setFrameworkListener(createFrameworkListener());
	}

	//
	// Instance Methods
	//

	private void checkShutdownIsAllowed(BundleContext bundleContext) {
		BundleContext bc = getBundleContext();
		if (bundleContext.equals(bc) == true) return;  // Early return.
		String message = Messages.getString(FrameworkManager.UNKNOWN_BUNDLE_CONTEXT_KEY);
		throw new IllegalArgumentException(message);
	}

	/**
	 * Create the manager's FrameworkListener.
	 *
	 * @return FrameworkListener.
	 */
	private FrameworkListener createFrameworkListener() {
		return new FrameworkListener() {
			public void frameworkEvent(FrameworkEvent event) {
				FrameworkManager.this.frameworkEvent(event);
			}
		};
	}

	/**
	 * Handle the specified framework event.
	 *
	 * @param event  A FrameworkEvent.
	 */
	private void frameworkEvent(FrameworkEvent event) {
		int type = event.getType();

		switch (type) {
			case FrameworkEvent.STARTED:
				handleFrameworkStarted(event);
				break;
			case FrameworkEvent.ERROR:
				handleFrameworkError(event);
				break;
			case FrameworkEvent.PACKAGES_REFRESHED:
				handleFrameworkPackagesRefreshed(event);
				break;
			case FrameworkEvent.STARTLEVEL_CHANGED:
				handleFrameworkStartLevelChanged(event);
				break;
			case FrameworkEvent.WARNING:
				handleFrameworkWarning(event);
				break;
			case FrameworkEvent.INFO:
				handleFrameworkInfo(event);
				break;
			default:
				handleUnknownFrameworkEvent(event);
				break;
		}
	}

	/**
	 * Request that garbage collection be performed.  The implementation of
	 * this method is attempting to be aggressive to encourage as much garbage
	 * as possible to be collected.  Of course calls to the <code>System</code>
	 * methods <code>gc()</code> and <code>runFinalization()</code> are never
	 * more than a hint to the VM.
	 */
	private void gc() {
		System.gc();
		System.gc();

		System.runFinalization();

		System.gc();
		System.gc();
	}

	/**
	 * Private <code>bundleContext</code> getter.
	 *
	 * @return BundleContext
	 */
	protected final BundleContext getBundleContext() {
		return bundleContext;
	}

	/**
	 * Private frameworkListener getter.
	 *
	 * @return The FrameworkListener.
	 */
	private FrameworkListener getFrameworkListener() {
		return frameworkListener;
	}

	/**
	 * Handle a framework error event.  This implementation simply logs the
	 * error to the <code>LogService</code>.
	 *
	 * @param event  The framework event.
	 */
	protected void handleFrameworkError(FrameworkEvent event) {
		logFrameworkError(event);
	}

	/**
	 * Handle a framework info event.  This implementation is a no-op.
	 *
	 * @param event  The framework event.
	 */
	protected void handleFrameworkInfo(FrameworkEvent event) {
		// No-op
	}

	/**
	 * Handle a framework packages refreshed event.  This implementation is a
	 * no-op.
	 *
	 * @param event  The framework event.
	 */
	protected void handleFrameworkPackagesRefreshed(FrameworkEvent event) {
		// No-op
	}

	/**
	 * Handle a framework started event.  This implementation simply logs the
	 * total memory, free memory and memory delta to the <code>LogService</code>.
	 *
	 * @param event  The framework event.
	 */
	protected void handleFrameworkStarted(FrameworkEvent event) {
		logMemoryStatistics();
	}

	/**
	 * Handle a framework start level changed event.  This implementation is a
	 * no-op.
	 *
	 * @param event  The framework event.
	 */
	protected void handleFrameworkStartLevelChanged(FrameworkEvent event) {
		// No-op
	}

	/**
	 * Handle a framework warning event.  This implementation is a no-op.
	 *
	 * @param event  The framework event.
	 */
	protected void handleFrameworkWarning(FrameworkEvent event) {
		// No-op
	}

	/**
	 * Handles the case where an unknown <code>FrameworkEvent</code> is
	 * received.
	 *
	 * @param event  The unknown <code>FrameworkEvent</code>.
	 */
	protected void handleUnknownFrameworkEvent(FrameworkEvent event) {
		logUnknownFrameworkEvent(event);
	}

	/**
	 * Hooks up the <code>FrameworkListener</code>.
	 */
	private void hookupFrameworkListener() {
		BundleContext bundleContext = getBundleContext();
		FrameworkListener listener = getFrameworkListener();
		bundleContext.addFrameworkListener(listener);
	}

	/**
	 * Answers true if the <code>BundleDependencyManager</code> is started,
	 * otherwise false.
	 *
	 * @return boolean
	 */
	public final boolean isStarted() {
		BundleContext bundleContext = getBundleContext();
		boolean started = bundleContext != null;
		return started;
	}

	/**
	 * Write a debug message to the <code>LogService</code>.
	 *
	 * @param message  The message to log.
	 */
	private void logDebug(String message) {
		LogUtility.logDebug(this, message);
	}

	/**
	 * Write an error message to the <code>LogService</code>.
	 *
	 * @param message  The message to log.
	 */
	private void logError(String message) {
		LogUtility.logError(this, message);
	}

	/**
	 * Write an error message and associated exception to the
	 * <code>LogService</code>.
	 *
	 * @param message    The message to log.
	 * @param throwable  The exception that was thrown and subsequently caught.
	 */
	private void logError(String message, Throwable throwable) {
		LogUtility.logError(this, message, throwable);
	}

	/**
	 * Log the framework event as an error.
	 *
	 * @param event  The framework event.
	 */
	protected final void logFrameworkError(FrameworkEvent event) {
		Throwable throwable = event.getThrowable();
		String message = Messages.getString(FrameworkManager.FRAMEWORK_ERROR_HAS_OCCURRED_KEY);
		logError(message, throwable);
	}

	/**
	 * Log the memory statistics.
	 */
	protected final void logMemoryStatistics() {
		Runtime runtime = Runtime.getRuntime();

		long total = runtime.totalMemory();
		long before = runtime.freeMemory();

		gc();

		long after = runtime.freeMemory();
		long delta = after - before;

		logMemoryStatistics(total, before, after, delta);
	}

	/**
	 * Log the specified memory statistics to the <code>LogService</code>.
	 *
	 * @param total   The total memory available to the VM.
	 * @param before  The memory consumed before a garbage collect.
	 * @param after   The memory consumed after a garbage collect.
	 * @param delta   The memory consumption delta.
	 */
	private void logMemoryStatistics(long total, long before, long after, long delta) {
		String pattern;
		String message;
		Long wrapper;

		pattern = Messages.getString(FrameworkManager.TOTAL_MEMORY_KEY);
		wrapper = new Long(total);
		message = MessageFormatter.format(pattern, wrapper);
		logDebug(message);

		pattern = Messages.getString(FrameworkManager.FREE_MEMORY_BEFORE_GC_KEY);
		wrapper = new Long(before);
		message = MessageFormatter.format(pattern, wrapper);
		logDebug(message);

		pattern = Messages.getString(FrameworkManager.FREE_MEMORY_AFTER_GC_KEY);
		wrapper = new Long(after);
		message = MessageFormatter.format(pattern, wrapper);
		logDebug(message);

		pattern = Messages.getString(FrameworkManager.FREE_MEMORY_DELTA_KEY);
		wrapper = new Long(delta);
		message = MessageFormatter.format(pattern, wrapper);
		logDebug(message);
	}

	/**
	 * Log the unknown framework event as an error.
	 *
	 * @param event  The framework event.
	 */
	protected final void logUnknownFrameworkEvent(FrameworkEvent event) {
		String pattern = Messages.getString(FrameworkManager.UNKNOWN_FRAMEWORK_EVENT_KEY);
		Object[] values = new Object [ 2 ];
		Class clazz = getClass();
		String name = clazz.getName();
		values [ 0 ] = name;
		values [ 1 ] = event;
		String message = MessageFormatter.format(pattern, values);
		logError(message);
	}

	/**
	 * Private <code>bundleContext</code> setter.
	 *
	 * @param bundleContext  The receiver's <code>BundleContext</code>.
	 */
	private void setBundleContext(BundleContext bundleContext) {
		if (this.bundleContext != null) {
			unhookFrameworkListener();
		}

		this.bundleContext = bundleContext;

		if (this.bundleContext != null) {
			hookupFrameworkListener();
		}
	}

	/**
	 * Private frameworkListener setter.
	 *
	 * @param frameworkListener  A FrameworkListener.
	 */
	private void setFrameworkListener(FrameworkListener frameworkListener) {
		this.frameworkListener = frameworkListener;
	}

	/**
	 * Shutdown the manager.
	 *
	 * @param bundleContext  The manager's <code>BundleContext</code>.
	 */
	public final void shutdown(BundleContext bundleContext) {
		Assert.checkArgumentIsNotNull(bundleContext, "bundleContext");  //$NON-NLS-1$

		synchronized (this) {
			boolean started = isStarted();
			if (started == false) return;  // Early return.
			checkShutdownIsAllowed(bundleContext);
			setBundleContext(null);
		}
	}

	/**
	 * Starts up the manager.
	 *
	 * @param bundleContext  The manager's <code>BundleContext</code>.
	 */
	public final void startup(BundleContext bundleContext) {
		Assert.checkArgumentIsNotNull(bundleContext, "bundleContext");  //$NON-NLS-1$

		synchronized (this) {
			boolean started = isStarted();
			if (started == true) return;  // Early return.
			setBundleContext(bundleContext);
		}
	}

	/**
	 * Unhook the <code>FrameworkListener</code>.
	 */
	private void unhookFrameworkListener() {
		BundleContext bundleContext = getBundleContext();
		FrameworkListener listener = getFrameworkListener();
		bundleContext.removeFrameworkListener(listener);
	}
}
